import { system, world, Player, MolangVariableMap } from "@minecraft/server";
import {applyKnockback} from "./relics"

const customProjectiles = new Set([
    "md_wardens:frostball_projectile",
    "md_wardens:fireball_projectile",
    "md_wardens:suspicious_pearl_projectile",
    "md_wardens:seismic_charge",
    "md_wardens:forest_projectile",
    "md_wardens:coral_projectile",
    "md_wardens:aoe_projectile"
]);

export const wardenSet = new Set ([
    "md_wardens:desert_warden",
    "md_wardens:forest_warden",
    "md_wardens:flower_warden",
    "md_wardens:cherry_blossom_warden",
    "md_wardens:mountain_warden",
    "md_wardens:savanna_warden",
    "md_wardens:snowy_warden",
    "md_wardens:jungle_warden",
    "md_wardens:ice_warden",
    "md_wardens:swamp_warden",
    "md_wardens:mesa_warden",
    "md_wardens:warm_ocean_warden",
    "md_wardens:mushroom_warden",
    "md_wardens:nether_warden",
    "md_wardens:ender_warden"
]);
// Map to store freeze intervals for each player
const freezeIntervals = new Map();
const cooldowns = new Map();
const COOLDOWN_DURATION = 5000;
let conduitInterval = null;
const trackedWardens = new Set();

// Allowed blocks set for root placement/replacement
const allowedBlocks = new Set([
    "minecraft:air",
    "minecraft:short_grass",
    "minecraft:tall_grass",
    "minecraft:vines",
    "minecraft:mangrove_roots",
    "minecraft:mangrove_leaves",
    "minecraft:water",
    "minecraft:lava",
    "minecraft:bamboo"
]);

world.afterEvents.projectileHitEntity.subscribe((event) => {
    const {dimension, projectile, location, source} = event;
    const hitEntity = event.getEntityHit().entity;
    const projectileType = projectile.typeId;

    if (source instanceof Player && wardenSet.has(hitEntity.typeId)) {
        const variantComponent = hitEntity.getComponent("minecraft:variant");
        if (variantComponent?.value === 12 || variantComponent?.value === 10) {
            const entityId = hitEntity.id;

            if (!checkAndSetCooldown(entityId)) {
                return; 
            }

            const playerLoc = source.location;
            const wardenLoc = hitEntity.location;

            const distance = Math.sqrt(
                Math.pow(playerLoc.x - wardenLoc.x, 2) +
                Math.pow(playerLoc.y - wardenLoc.y, 2) +
                Math.pow(playerLoc.z - wardenLoc.z, 2)
            );

            if (distance > 16) {
                if (variantComponent.value === 12) {
                    startSnowStormAOE(source);
                } else if (variantComponent.value === 10) {
                    startsporeAOE(source);
                }
            } else {
                cooldowns.delete(entityId);
            }

            system.runTimeout(() => {
                cooldowns.delete(entityId);
            }, COOLDOWN_DURATION);
        }
    }

    // Early exit if the projectile is not one of the custom types
    if (!customProjectiles.has(projectileType)) {
        return;
    }

    switch (projectileType) {
        case "md_wardens:frostball_projectile":
            freezeEntityInIce(hitEntity);
            source.triggerEvent('switch_to_melee')
            break;
        case "md_wardens:suspicious_pearl_projectile":
            teleportEntityRandomly(hitEntity, dimension, location);
            source.triggerEvent('switch_to_melee')
            break;
    }
});

world.afterEvents.worldInitialize.subscribe(() => {
    loadTrackedWardens();
    system.runTimeout(() => {
        trackConduit();
    },200)
});

world.afterEvents.entityDie.subscribe((e) => {
    const ent = e.deadEntity;
    if (wardenSet.has(ent.typeId)) {
        trackedWardens.delete(ent.id);
        saveTrackedWardens();
    }
    // Clear freeze interval for the dead entity
    clearFreezeInterval(ent.id);
});

world.afterEvents.entitySpawn.subscribe((e) => {
  const { entity } = e;
    if (!wardenSet.has(entity.typeId) && !customProjectiles.has(entity.typeId)) {
        return;
    }

    if (wardenSet.has(entity.typeId)) {
        const variant = entity.getComponent("minecraft:variant");
        if (variant.value === 13) {
            trackedWardens.add(entity.id);
            saveTrackedWardens();
            trackConduit();
        }
    }


  const dim = entity.dimension;
  const nearbyWardens = dim.getEntities({
    location: entity.location,
    maxDistance: 3
  }).filter(entity => 
    wardenSet.has(entity.typeId)
);

    if (nearbyWardens.length > 0) {
        const warden = nearbyWardens[0]; // Assume the closest warden is the one who shot the projectile
        const variantComponent = warden.getComponent("minecraft:variant");
        if (!variantComponent) return;

        if (variantComponent.value === 14 && customProjectiles.has(entity.typeId)) {
            boulder(warden, entity)
        }
        
        if (entity.typeId === "md_wardens:aoe_projectile") {
            switch (variantComponent.value) {
                case 12: 
                startSnowStormAOE(warden);
                    break;
                case 8: 
                earthquake(warden);
                    break;
                case 1: 
                cherryBoom(warden);
                    break;
                case 4: 
                spawnFlowerFangs(warden, warden.location);
                    break;
                case 10: 
                    startsporeAOE(warden);
                    break;
                case 7: 
                    spawnBambooPath(warden, warden.location);
                    break;
                case 15: 
                    spawnRootWave(warden, warden.location);
                    break;
                case 5:
                    forestBoom(entity)
                    break;
                case 11:
                    stomp(warden)
                    break;
            }
            warden.triggerEvent('switch_to_melee');
            entity.remove();
        } else if ((variantComponent.value === 9 || variantComponent.value === 13) && customProjectiles.has(entity.typeId)) {
            warden.triggerEvent('switch_to_melee');
        }
    }
});

function stomp(warden) {
    const dim = warden.dimension;
    const loc = warden.location;
    const targetTypes = ["minecraft:player", "minecraft:villager", "minecraft:wandering_trader"];
    const molangVars = new MolangVariableMap();
    // Set random values
    molangVars.setFloat("variable.dustsize", Math.random(10,15));
    molangVars.setFloat("variable.particle_random_n", Math.random(1,10));
    dim.spawnParticle("md_warden:dustcloud", {x: loc.x, y: loc.y +0.2, z: loc.z}, molangVars);

    
    system.runTimeout(() => {
        warden.triggerEvent('switch_to_slow_melee');
    }, 60);

    // Find the nearest targets and send them upward a bit
    const potentialTargets = dim.getEntities({
        location: loc,
        maxDistance: 30 
    }).filter(target => targetTypes.includes(target.typeId));

    for (const target of potentialTargets) {
        applyKnockback(target, loc, 0, 0.5)
        target.applyDamage(3)
        if (target instanceof Player) {
            target.runCommand("camerashake add @p 0.5 1.5 positional");
            target.playSound("maca.stomp1")
        }
    }    
}

function trackConduit() {
    if (conduitInterval === null) {
        conduitInterval = system.runInterval(() => {
            for (const wardenId of trackedWardens) {
                const warden = world.getEntity(wardenId);
                if (warden) {
                    conduitPower(warden);
                } else {
                    trackedWardens.delete(wardenId);
                    saveTrackedWardens();
                }
            }
        }, 40);
    }
} 

function saveTrackedWardens() {
  const wardenIds = Array.from(trackedWardens);
  world.setDynamicProperty("trackedWardens", JSON.stringify(wardenIds));
}

function loadTrackedWardens() {
  const savedWardens = world.getDynamicProperty("trackedWardens");
  if (savedWardens) {
    const wardenIds = JSON.parse(savedWardens);
    trackedWardens.clear();
    wardenIds.forEach(id => trackedWardens.add(id));
  }
}

function boulder(entity, projectile) {
    const targetTypes = ["minecraft:player", "minecraft:villager", "minecraft:wandering_trader"];
    let nearestTarget = null;
    let shortestDistance = Infinity;

    // Find the nearest target
    const potentialTargets = entity.dimension.getEntities({
        location: entity.location,
        maxDistance: 60 // Adjust this range as needed
    }).filter(target => targetTypes.includes(target.typeId));

    for (const target of potentialTargets) {
        const distance = calculateDistance(entity.location, target.location);
        if (distance < shortestDistance) {
            shortestDistance = distance;
            nearestTarget = target;
        }
    }

    // If no target found, use entity's view direction
    let targetLocation = nearestTarget ? nearestTarget.location : null;
    if (!targetLocation) {
        const viewDirection = entity.getViewDirection();
        targetLocation = {
            x: entity.location.x + viewDirection.x * 20,
            y: entity.location.y + viewDirection.y * 20,
            z: entity.location.z + viewDirection.z * 20
        };
    }

    // Initial position
    let startPos = {
        x: entity.location.x,
        y: entity.location.y + 4, // Adjust based on where the boulder should start relative to the entity
        z: entity.location.z
    };

    // Set initial position of the projectile
    projectile.teleport(startPos);

    // Hold the projectile for 1 second
    let holdTimer = 0;
    const holdInterval = system.runInterval(() => {
        projectile.applyImpulse({ x: 0, y: 0.05, z: 0 }); // Small upward impulse to hold it
        holdTimer++;
        if (holdTimer >= 30) {
            system.clearRun(holdInterval);
            launchBoulder();
        }
    }, 1);

    function launchBoulder() {
        const targetPos = nearestTarget ? nearestTarget.location : targetLocation;
        const direction = {
            x: targetPos.x - startPos.x,
            y: targetPos.y - startPos.y,
            z: targetPos.z - startPos.z
        };
        const horizontalDistance = Math.sqrt(direction.x * direction.x + direction.z * direction.z);
        const totalDistance = Math.sqrt(horizontalDistance * horizontalDistance + direction.y * direction.y);
    
        // Adjust flight time based on distance
        const totalFlightTime = Math.max(20, Math.min(60, Math.floor(totalDistance / 1.5)));
    
        // Calculate a higher arc than before to avoid it dropping on the wardens head
        const maxArcHeight = Math.max(10, totalDistance / 3);
    
        // Calculate initial velocity
        const initialVelocity = totalDistance / totalFlightTime;
    
        system.runTimeout(() => {
            entity.triggerEvent('switch_to_melee');
        }, 10);
    
        let currentTick = 0;
        const flightInterval = system.runInterval(() => {
            currentTick++;
            if (!projectile.isValid()) {
                system.clearRun(flightInterval);
                return;
            }
    
            const progress = currentTick / totalFlightTime;
            const verticalProgress = 4 * progress * (1 - progress);
    
            // Calculate the current desired position
            const desiredPos = {
                x: startPos.x + direction.x * progress,
                y: startPos.y + direction.y * progress + maxArcHeight * verticalProgress,
                z: startPos.z + direction.z * progress
            };
    
            // Calculate the needed velocity to reach the desired position
            const currentPos = projectile.location;
            const neededVelocity = {
                x: (desiredPos.x - currentPos.x) * 2,
                y: (desiredPos.y - currentPos.y) * 2,
                z: (desiredPos.z - currentPos.z) * 2
            };
    
            // Scale the velocity based on the initial velocity calculation
            const velocityScale = initialVelocity / 2; // Adjust this divisor to control overall speed
            neededVelocity.x *= velocityScale;
            neededVelocity.y *= velocityScale;
            neededVelocity.z *= velocityScale;
    
            // Apply the calculated impulse
            projectile.applyImpulse(neededVelocity);
    
            // Check if we've reached the end of the flight or hit the target
            if (currentTick >= totalFlightTime || calculateDistance(projectile.location, targetPos) < 0.5) {
                system.clearRun(flightInterval);
            }
        }, 1);
    }
}

// Helper function to calculate distance between two points
function calculateDistance(pos1, pos2) {
    const dx = pos2.x - pos1.x;
    const dy = pos2.y - pos1.y;
    const dz = pos2.z - pos1.z;
    return Math.sqrt(dx * dx + dy * dy + dz * dz);
}

function spawnFlowerFangs(warden, startLoc) {
    const dim = warden.dimension;
    const delayBetweenSpawns = 2; // Ticks between each spawn
    const branchCount = 3; // Number of branches


    const targetTypes = ["minecraft:player"];
    let maxDistance = 15;
    let targetPlayer = null;

    // Find the nearest target
    const potentialTargets = dim.getEntities({
        location: warden.location,
        maxDistance: 35 // Adjust this range as needed
    }).filter(target => targetTypes.includes(target.typeId));

    if (potentialTargets.length > 0) {
        targetPlayer = potentialTargets[0]; // Assuming the first player is the target
        maxDistance = calculateDistance(warden.location, targetPlayer.location) + 5;
        dim.spawnParticle("md_wardens:sun_ray", targetPlayer.location);

        // Calculate direction to the player
        const playerDirection = {
            x: targetPlayer.location.x - startLoc.x,
            z: targetPlayer.location.z - startLoc.z
        };
        const playerDirectionLength = Math.sqrt(playerDirection.x * playerDirection.x + playerDirection.z * playerDirection.z);
        const normalizedPlayerDirection = {
            x: playerDirection.x / playerDirectionLength,
            z: playerDirection.z / playerDirectionLength
        };

        // Spawn the main branch targeting the player
        spawnBranch(normalizedPlayerDirection, 0);

        // Spawn additional branches
        for (let branch = 1; branch < branchCount; branch++) {
            const branchAngle = (Math.random() - 0.5) * Math.PI / 2; // Random angle between -π/4 and π/4
            const branchDirection = {
                x: normalizedPlayerDirection.x * Math.cos(branchAngle) - normalizedPlayerDirection.z * Math.sin(branchAngle),
                z: normalizedPlayerDirection.x * Math.sin(branchAngle) + normalizedPlayerDirection.z * Math.cos(branchAngle)
            };

            // Random start point along the main branch
            const startOffset = Math.floor(Math.random() * (maxDistance / 2));
            spawnBranch(branchDirection, startOffset);
        }

        system.runTimeout(()=>{
            // Start damage application
        applyDamageOverTime(targetPlayer.location);
        },30)
    }

    function spawnBranch(direction, startOffset) {
        let currentTick = startOffset;

        system.runInterval(() => {
            if (currentTick >= maxDistance) {
                return "stop";
            }

            const spawnX = Math.round(startLoc.x + direction.x * currentTick);
            const spawnZ = Math.round(startLoc.z + direction.z * currentTick);

            // Cast a ray to find the ground level
            const raycastOptions = {
                maxDistance: 10,
                includeLiquidBlocks: false,
                includePassableBlocks: false
            };
            const raycastResult = dim.getBlockFromRay({ x: spawnX, y: startLoc.y + 5, z: spawnZ }, { x: 0, y: -1, z: 0 }, raycastOptions);

            if (raycastResult) {
                const spawnY = raycastResult.block.y + 1; // Spawn one block above the found ground

                dim.spawnEntity("md_wardens:flower_entity_projectile", { x: spawnX, y: spawnY, z: spawnZ });
            }

            currentTick++;
        }, delayBetweenSpawns);
    }

    function applyDamageOverTime(centerLocation) {
        const damageRadius = 8; // Radius for damage application
        const damageAmount = 2; // Amount of damage to apply each second
        const attackDuration = 4.5; // Duration of the attack in seconds
        const fireDuration = 2; // Duration of fire effect in seconds
        let hitCount = 0;
        const damageInterval = system.runInterval(() => {
            hitCount++;
            const entities = dim.getEntities({
                location: centerLocation,
                maxDistance: damageRadius
            });
    
            entities.forEach(entity => {
                if (!wardenSet.has(entity.typeId) && 
                    entity.typeId !== "minecraft:item" && 
                    entity.typeId !== "minecraft:xp_orb" && 
                    entity.typeId !== "md_wardens:flower_entity_projectile") {
                    
                    // Apply damage
                    entity.applyDamage(damageAmount);
    
                    // Set on fire if it's the second hit or later
                    if (hitCount >= 3) {
                        entity.setOnFire(fireDuration, true);
                    }
                }
            });
    
            // Stop the interval after the attack duration
            if (hitCount >= attackDuration) {
                system.clearRun(damageInterval);
            }
        }, 20); // Run every second (20 ticks)
    }
}

function spawnRootWave(warden, startLoc) {
    const dim = warden.dimension;
    const delayBetweenSpawns = 1;
    const targetTypes = ["minecraft:player"];
    let maxDistance = 20;
    let targetPlayer = null;

    const waveAmplitude = 2;
    const waveFrequency = 0.4;

    const potentialTargets = dim.getEntities({
        location: warden.location,
        maxDistance: 30
    }).filter(target => targetTypes.includes(target.typeId));

    if (potentialTargets.length > 0) {
        targetPlayer = potentialTargets[0];
        maxDistance = calculateDistance(warden.location, targetPlayer.location) + 5;
        placeRootBlocks();
    }

    function placeRootBlocks() {
        let currentTick = 0;
        const waveLength = 8;
    
        const rootInterval = system.runInterval(() => {
            if (currentTick >= maxDistance) {
                system.clearRun(rootInterval);
                return;
            }

            const playerDirection = {
                x: targetPlayer.location.x - startLoc.x,
                z: targetPlayer.location.z - startLoc.z
            };
            const playerDirectionLength = Math.sqrt(playerDirection.x * playerDirection.x + playerDirection.z * playerDirection.z);
            const normalizedPlayerDirection = {
                x: playerDirection.x / playerDirectionLength,
                z: playerDirection.z / playerDirectionLength
            };
    
            const spawnX = Math.round(startLoc.x + normalizedPlayerDirection.x * currentTick);
            const spawnZ = Math.round(startLoc.z + normalizedPlayerDirection.z * currentTick);
    
            const raycastOptions = {
                maxDistance: 40,
                includeLiquidBlocks: false,
                includePassableBlocks: false
            };
            const raycastResult = dim.getBlockFromRay(
                { x: spawnX, y: startLoc.y + 10, z: spawnZ },
                { x: 0, y: -1, z: 0 },
                raycastOptions
            );
    
            if (raycastResult) {
                const groundY = raycastResult.block.y + 1;
                const normalizedDistance = (currentTick % waveLength) / waveLength;
                const waveOffset = Math.sin(normalizedDistance * 2 * Math.PI) * waveAmplitude;
                const spawnY = Math.round(groundY + waveOffset);
                const newLoc = { x: spawnX, y: spawnY, z: spawnZ };
                const blockBelowId = dim.getBlock(newLoc);
                
                if (allowedBlocks.has(blockBelowId.typeId)) {
                    dim.setBlockType(newLoc, "minecraft:mangrove_roots");
    
                    system.runTimeout(() => {
                        dim.setBlockType(newLoc, "minecraft:air");
                    }, 10);
    
                    checkForEntities(spawnX, spawnY, spawnZ);
                }
            }
    
            currentTick++;
        }, delayBetweenSpawns);
    }

    function checkForEntities(x, y, z) {
        const damageRadius = 5;
        const entitiesInRange = dim.getEntities({
            location: { x: x, y: y, z: z },
            maxDistance: damageRadius
        });

        entitiesInRange.forEach(entity => {
            if (targetTypes.includes(entity.typeId)) {
                entity.applyDamage(6);
            }
        });
    }
}

function spawnBambooPath(warden, startLoc) {
    const dim = warden.dimension;
    const delayBetweenSpawns = 2; // Ticks between each block placement
    const targetTypes = ["minecraft:player"];
    let maxDistance = 20; // Maximum distance for bamboo placement
    let targetPlayer = null;

    // Find the nearest target
    const potentialTargets = dim.getEntities({
        location: warden.location,
        maxDistance: 20
    }).filter(target => targetTypes.includes(target.typeId));

    if (potentialTargets.length > 0) {
        targetPlayer = potentialTargets[0]; // Assuming the first cow is the target
        maxDistance = calculateDistance(warden.location, targetPlayer.location) + 5;

        // Calculate direction to the player
        const playerDirection = {
            x: targetPlayer.location.x - startLoc.x,
            z: targetPlayer.location.z - startLoc.z
        };
        const playerDirectionLength = Math.sqrt(playerDirection.x * playerDirection.x + playerDirection.z * playerDirection.z);
        const normalizedPlayerDirection = {
            x: playerDirection.x / playerDirectionLength,
            z: playerDirection.z / playerDirectionLength
        };

        // Start placing bamboo along the path
        placeBambooPath(normalizedPlayerDirection);
    }

    function placeBambooPath(direction) {
        let currentTick = 0;

        system.runInterval(() => {
            if (currentTick >= maxDistance) {
                return "stop";
            }

            // Calculate positions for bamboo placement in a 3-block wide path
            for (let offsetX = -1; offsetX <= 1; offsetX++) {
                const spawnX = Math.round(startLoc.x + direction.x * currentTick + offsetX);
                const spawnZ = Math.round(startLoc.z + direction.z * currentTick);

                // Cast a ray to find the ground level for proper placement
                const raycastOptions = {
                    maxDistance: 40,
                    includeLiquidBlocks: false,
                    includePassableBlocks: false
                };
                const raycastResult = dim.getBlockFromRay(
                    { x: spawnX, y: startLoc.y + 10, z: spawnZ }, // Start higher to account for hills
                    { x: 0, y: -1, z: 0 },
                    raycastOptions
                );

                if (raycastResult) {
                    const groundY = raycastResult.block.y + 1; // One block above ground level

                    // Random height for bamboo (1 to 3 blocks high)
                    const bambooHeight = Math.floor(Math.random() * 15) + 1;
                    const newLoc = { x: spawnX, y: groundY, z: spawnZ }
                    const blockBelowId = dim.getBlock(newLoc);
                    if (allowedBlocks.has(blockBelowId.typeId)) {
                        // Check for entities in the area and apply damage and knockback if necessary
                        checkForEntities(spawnX, groundY, spawnZ);
                        system.runTimeout(()=>{
                            // Place bamboo blocks vertically at the calculated position
                            for (let h = 0; h < bambooHeight; h++) {
                                dim.setBlockType({ x: spawnX, y: groundY + h, z: spawnZ }, "minecraft:bamboo");
                            }
                        },2)
                        }
                } 
            }

            currentTick++; // Increment consistently for each tick to place blocks along the path
        }, delayBetweenSpawns);
    }

    function checkForEntities(x, y, z) {
        const damageRadius = 3; // Radius to check for entities around the placed block
        const entitiesInRange = dim.getEntities({
            location: { x: x, y: y, z: z },
            maxDistance: damageRadius
        });

        entitiesInRange.forEach(entity => {
            if (!wardenSet.has(entity.typeId) && entity.typeId !== "minecraft:item" && entity.typeId !== "minecraft:xp_orb") {
                entity.applyDamage(2); // Apply damage to entities within range
                // Define knockback strength
                const knockbackStrength = 2;
                // Apply knockback to the entity using only vertical strength
                entity.applyKnockback(
                    0,
                    0,
                    knockbackStrength,
                    knockbackStrength
                );
            }
        });
    }
}

// Function to check and set cooldown
export function checkAndSetCooldown(entityId) {
    const currentTime = Date.now();
    if (cooldowns.has(entityId)) {
        const lastUsedTime = cooldowns.get(entityId);
        const timeRemaining = COOLDOWN_DURATION - (currentTime - lastUsedTime);
        if (timeRemaining > 0) {
            return false;
        }
    }
    cooldowns.set(entityId, currentTime);
    return true;
}

function conduitPower(entity) {
    const loc = entity.location;
    const dim = entity.dimension;
    const radius = 23;

    const players = dim.getPlayers({
        location: loc,
        maxDistance: radius
    });

    players.forEach(player => {
        player.addEffect('conduit_power', 300, { amplifier: 0, showParticles: false });
    });
}

function forestBoom(warden) {
    const dim = warden.dimension;
    const loc = warden.location;
    const dir = warden.getViewDirection();

    const targetLocation = {
        x: loc.x ,
        y: loc.y + 1.5,
        z: loc.z
    };

    dim.spawnParticle("md_wardens:forest_sonic_explosion", targetLocation);
    system.runTimeout(() => {
        sonicAOE(dim, loc)
        dim.playSound("mob.warden.sonic_boom",loc)
    }, 10);
}

function cherryBoom(warden) {
  const dim = warden.dimension;
  const loc = warden.location;
  const dir = warden.getViewDirection();

  const targetLocation = {
    x: loc.x ,
    y: loc.y + 1,
    z: loc.z
};
  
  // Normalize the direction vector to ensure values are between -1 and 1
  const magnitude = Math.sqrt(dir.x * dir.x + dir.z * dir.z);
  const dirX = dir.x / magnitude;
  const dirZ = dir.z / magnitude;

  const molangVars = new MolangVariableMap();
  molangVars.setFloat("variable.direction_x", dirX);
  molangVars.setFloat("variable.direction_z", dirZ);
  dim.spawnParticle("md_wardens:cherry_charge", targetLocation);
  
  system.runTimeout(() => {
    dim.spawnParticle(
        "md_wardens:cherry_sonic_explosion", 
        targetLocation,
        molangVars
      );
      dim.playSound("mob.warden.sonic_boom",loc)
    sonicAOE(dim, loc);
  }, 5);

  system.runTimeout(() => {
    dim.spawnParticle("md_wardens:cherry_explosion", targetLocation);
    dim.spawnParticle("md_wardens:cherry_explosion_disc", targetLocation);
  }, 10);
}

function sonicAOE(dim, loc) {
    const radius = 30; 
    const knockbackStrength = 4.5; 

    const entities = dim.getEntities({
        location: loc,
        maxDistance: radius
    }).filter(entity => 
        entity.typeId !== "minecraft:item" && 
        !wardenSet.has(entity.typeId) && 
        entity.typeId !== "md_wardens:aoe_projectile" && 
        entity.typeId !== "minecraft:xp_orb"
    );

    entities.forEach(entity => {
        // Calculate direction vector from blast center to entity
        const dx = entity.location.x - loc.x;
        const dy = entity.location.y - loc.y;
        const dz = entity.location.z - loc.z;

        // Calculate distance
        const distance = Math.sqrt(dx * dx + dy * dy + dz * dz);

        // Normalize the direction vector
        const nx = dx / distance;
        const ny = dy / distance;
        const nz = dz / distance;

        // Calculate knockback velocity
        const knockbackVelocityX = nx * knockbackStrength;
        const knockbackVelocityY = ny * knockbackStrength;
        const knockbackVelocityZ = nz * knockbackStrength;

        // Apply knockback to the entity
        entity.applyKnockback(
            knockbackVelocityX,
            knockbackVelocityZ,
            knockbackStrength,
            knockbackVelocityY
        );

        entity.applyDamage(5); 
    });
}

function earthquake(warden) {
    const loc = warden.location;
    const dim = warden.dimension;

    const radius = 20;

    // Play sound effect
    dim.playSound("mob.warden.roar", loc);

    // Get nearby entities and apply effects
    const entities = dim.getEntities({
        location: loc,
        maxDistance: radius
    }).filter(entity =>
        entity.typeId !== "minecraft:item" &&
        !wardenSet.has(entity.typeId) &&
        entity.typeId !== "md_wardens:aoe_projectile" &&
        entity.typeId !== "minecraft:xp_orb"
    ).forEach(entity => {
        if (entity instanceof Player) {
            entity.runCommand("camerashake add @p 0.5 5 positional");
            
            // Calculate direction from warden to player
            const directionToPlayer = {
                x: entity.location.x - warden.location.x,
                z: entity.location.z - warden.location.z
            };
            
            // Normalize the direction
            const length = Math.sqrt(directionToPlayer.x * directionToPlayer.x + directionToPlayer.z * directionToPlayer.z);
            const normalizedDirection = {
                x: directionToPlayer.x / length,
                z: directionToPlayer.z / length
            };
            
            // Calculate fissure start location (7 blocks away from player in the direction they're running)
            const fissureStartLoc = {
                x: entity.location.x + normalizedDirection.x * 8,
                y: entity.location.y,
                z: entity.location.z + normalizedDirection.z * 8
            };
            
            createFissure(dim, fissureStartLoc);
        }
        entity.addEffect("slowness", 100, { amplifier: 3 });

        for (let i = 0; i < 4; i++) {
            system.runTimeout(() => {
                try {
                    entity.applyDamage(1.5);
                } catch (error) {
                }
            }, 20 * i);
        }
    });
}

function createFissure(dim, loc) {
    // Define fissure parameters
    const fissureLength = Math.floor(Math.random() * (15 - 10 + 1)) + 10; // Random length between 10 and 15 blocks
    const fissureDepth = Math.floor(Math.random() * (5 - 3 + 1)) + 3; // Random depth between 3 and 5 blocks

    const blocksToDestroy = [];
    for (let lengthOffset = -Math.floor(fissureLength / 2); lengthOffset <= Math.floor(fissureLength / 2); lengthOffset++) {
        // Calculate current Z position based on length offset
        const currentZ = loc.z + lengthOffset;

        // Calculate dynamic width based on distance from center
        const maxWidthAtStart = Math.floor(Math.random() * (6 - 4 + 1)) + 4; // Random initial width between 3 and 5 blocks
        const currentWidth = Math.max(0, maxWidthAtStart - Math.abs(lengthOffset)); // Decrease width with distance

        for (let widthOffset = -Math.floor(currentWidth / 2); widthOffset <= Math.floor(currentWidth / 2); widthOffset++) {
            // Calculate X position based on width offset
            const currentX = loc.x + widthOffset;

            // Raycast for each position to find the Y level of the top layer
            const raycastOptions = {
                maxDistance: 40,
                includeLiquidBlocks: false,
                includePassableBlocks: false
            };

            const raycastResult = dim.getBlockFromRay(
                { x: currentX, y: loc.y + 5, z: currentZ }, // Start higher to account for hills
                { x: 0, y: -1, z: 0 },
                raycastOptions
            );

            if (raycastResult) {
                const groundY = raycastResult.block.y; // Ground level found

                // Add blocks for each depth level along the fissure's width and length
                for (let depth = 0; depth < fissureDepth; depth++) {
                    const currentY = groundY - depth; // Calculate Y position for lower layers

                    // Calculate effective length at this depth (decrease by depth)
                    const effectiveLength = fissureLength - depth * 2;

                    if (lengthOffset >= -Math.floor(effectiveLength / 2) && lengthOffset <= Math.floor(effectiveLength / 2)) {
                        blocksToDestroy.push({ x: currentX, y: currentY, z: currentZ });

                        // Spawn particle at the top layer position
                        if (depth === 0) {
                            dim.spawnParticle("md_wardens:block_destruct", { x: currentX, y: groundY +1, z: currentZ });
                        }
                    }
                }
            } 
        }
    }

    // Remove blocks just after the particle animation finishes
    system.runTimeout(() => {
        blocksToDestroy.forEach(block => {
            dim.setBlockType(block, "minecraft:air");
        });
    }, 40);
}

export function startSnowStormAOE(warden) {
  let intervalCount = 0;
  const maxIntervals = 3;
  const dim = warden.dimension;
  const loc = warden.location;
  dim.playSound("mob.warden.roar", loc)
  dim.spawnParticle("md_wardens:encircle", loc);
  dim.spawnParticle("md_wardens:snow", loc);

  const intervalId = system.runInterval(() => {
    snowStormAOE(warden);
    intervalCount++;

    if (intervalCount === maxIntervals) {
      system.clearRun(intervalId);
    }
  }, 100); // 100 ticks = 5 seconds
}

function snowStormAOE(warden) {
  const projectilePos = warden?.location;
  const dim = warden?.dimension;
  freezeEntitiesInRadius(projectilePos, dim, warden);
}

function freezeEntitiesInRadius(projectilePosition, dim, warden) {
  const radius = 15; // 15-block radius

  const entities = dim.getEntities({
    location: projectilePosition,
    maxDistance: radius
  }).filter(entity => 
    entity.typeId !== "minecraft:item" && 
    entity.typeId !== warden.typeId && 
    entity.typeId !== "md_wardens:aoe_projectile" && 
    entity.typeId !== "minecraft:xp_orb"
  );

  entities.forEach(entity => {
    entity.addEffect("slowness", 200, { amplifier: 3 });
    simulateFreezingDamage(entity);
  });
}

function simulateFreezingDamage(entity) {
  let freezeTicks = 0;
  const maxFreezeTicks = 140; // Approximately 7 seconds of freezing
  const damageInterval = 40; // Apply damage every 2 seconds

  const freezeIntervalId = system.runInterval(() => {
    if (!entity.isValid() || freezeTicks >= maxFreezeTicks) {
      clearFreezeInterval(entity.id);
      return;
    }

    freezeTicks++;

    if (freezeTicks % damageInterval === 0) {
      entity.applyDamage(0.5, { cause: "freezing" });
    }
  }, 1);

  freezeIntervals.set(entity.id, freezeIntervalId);
}

function clearFreezeInterval(entityId) {
  const intervalId = freezeIntervals.get(entityId);
  if (intervalId) {
    system.clearRun(intervalId);
    freezeIntervals.delete(entityId);
  }
}

// Function to freeze an entity in a 3x3x3 cube of frosted ice
function freezeEntityInIce(entity) {
    const entityLocation = entity.location;
    const dimension = entity.dimension;

    const range = [-1, 0, 1];
    range.forEach((dx) => {
        range.forEach((dy) => {
            range.forEach((dz) => {
                const blockLocation = {
                    x: Math.floor(entityLocation.x) + dx,
                    y: Math.floor(entityLocation.y) + dy +1,
                    z: Math.floor(entityLocation.z) + dz,
                };

                const newBlock = dimension.getBlock(blockLocation);
                if (newBlock && newBlock.typeId === "minecraft:air") {
                    newBlock.setType("minecraft:powder_snow");
                }
            });
        });
    });
}

// Function to teleport an entity to a random location within a 20x20x20 cube
function teleportEntityRandomly(entity, dimension, originalLocation) {
    const maxAttempts = 16;
    const cubeSize = 20;

    function attemptTeleport(attempt) {
        if (attempt >= maxAttempts) {
            return;
        }

        // Generate a random point within the cube
        const offsetX = (Math.random() - 0.5) * cubeSize;
        const offsetY = (Math.random() - 0.5) * cubeSize;
        const offsetZ = (Math.random() - 0.5) * cubeSize;

        const targetX = originalLocation.x + offsetX;
        const targetY = originalLocation.y + offsetY;
        const targetZ = originalLocation.z + offsetZ;

        // Perform a raycast from the target point downwards
        const raycastResult = dimension.getBlockFromRay(
            { x: targetX, y: targetY, z: targetZ },
            { x: 0, y: -1, z: 0 },
            {
                maxDistance: cubeSize,
                includeLiquidBlocks: false,
                includePassableBlocks: false
            }
        );

        if (raycastResult) {
            const landingBlock = raycastResult.block;
            const teleportLocation = {
                x: landingBlock.location.x + 0.5,
                y: landingBlock.location.y + 1,
                z: landingBlock.location.z + 0.5
            };

            // Check if there's enough headroom (2 blocks of air)
            const headBlock1 = dimension.getBlock({x: teleportLocation.x, y: teleportLocation.y, z: teleportLocation.z});
            const headBlock2 = dimension.getBlock({x: teleportLocation.x, y: teleportLocation.y + 1, z: teleportLocation.z});

            if (headBlock1.typeId === "minecraft:air" && headBlock2.typeId === "minecraft:air") {
                // Teleport the entity
                entity.teleport(
                    teleportLocation,
                    {
                        dimension: dimension,
                        rotation: { x: 0, y: 0 },
                        keepVelocity: false
                    }
                );
            } else {
                // If not enough headroom, try again
                system.run(() => attemptTeleport(attempt + 1));
            }
        } else {
            // If no suitable block found, try again
            system.run(() => attemptTeleport(attempt + 1));
        }
    }

    // Start the teleportation attempts
    attemptTeleport(0);
}

function startsporeAOE(warden) {
  const dim = warden.dimension;
  const loc = warden.location;
  dim.playSound("mob.warden.roar", loc)
  dim.spawnParticle("md_wardens:encircle", loc);
  dim.spawnParticle("md_wardens:spores", loc);
  poisonEntitiesInRadius(loc, dim);
}

function poisonEntitiesInRadius(projectilePosition, dim) {
  const radius = 20;
  dim.getEntities({
    location: projectilePosition,
    maxDistance: radius
  }).filter(entity => 
    entity.typeId !== "minecraft:item" && 
    !wardenSet.has(entity.typeId) && 
    entity.typeId !== "md_wardens:aoe_projectile" && 
    entity.typeId !== "minecraft:xp_orb"
  ).forEach(entity => {
    entity.addEffect("poison", 200);
  });
}
